##############################################################################
#
# Copyright (c) 2001, 2002, 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""$Id: OODocument.py,v 1.16 2003/03/16 18:15:17 longsleep Exp $
"""

# html rendered OpenOffice Documents for CMF/Plone by Simon Eisenmann

import Globals, string, re, os, sys
from Globals import DTMLFile, InitializeClass
from StringIO import StringIO
from AccessControl import ClassSecurityInfo, getSecurityManager

import Acquisition
from Acquisition import aq_base, aq_parent
from OFS.Traversable import Traversable
from webdav.Collection import Collection
from ExtensionClass import Base
from OFS.content_types import guess_content_type

import OFS.Image
from BTrees.OOBTree import OOBTree

from Products.CMFCore.PortalContent import NoWL, ResourceLockedError
from Products.CMFCore.CMFCorePermissions import View
from Products.CMFCore.CMFCorePermissions import ModifyPortalContent
from Products.CMFCore.WorkflowCore import WorkflowAction
from Products.CMFCore.utils import keywordsplitter, _verifyActionPermissions, getToolByName
from Products.CMFDefault import utils
from Products.CMFDefault.Document import Document
from Products.CMFDefault.File import File
from Products.CMFDefault.DublinCore import DefaultDublinCoreImpl
from Products.CMFDefault.utils import parseHeadersBody, formatRFC822Headers
from Products.CMFDefault.utils import SimpleHTMLParser, bodyfinder, _dtmldir
from DocumentTemplate.DT_Util import html_quote

import OFS.Image

import libxml2
import libxslt
import zipfile
import tempfile

try: 
    import PIL.Image
    isPilAvailable = 1
except ImportError: 
    isPilAvailable = 0

_DEBUG_XSLT = 0
text_format = 'application/vnd.sun.xml.writer'

# deactivate libxml2 and libxslt error messages
if not _DEBUG_XSLT:
    def noerr(ctx, str):
        pass
    libxml2.registerErrorHandler(noerr, None)
    libxslt.registerErrorHandler(noerr, None)

# load the openoffice 2 html transformation stylesheet
packageDir=Globals.package_home(globals())
__stylefile=packageDir.split(os.sep)+['sx2ml', 'main_html.xsl']
__stylefile='file:///'+'/'.join(__stylefile)
__styledoc=libxml2.parseFile(__stylefile)
_style=libxslt.parseStylesheetDoc(__styledoc)

# fill new oodocument instances with an empty openoffice document
__emptyoodocfile=packageDir.split(os.sep)+['_EMPTY.sxw']
__emptyoodocfile='/'.join(__emptyoodocfile)
def loadEmptyOODocument():

    try:
	try:
            doc = open(__emptyoodocfile, 'rb', -1)
            return doc.read()
	finally:
            doc.close()
    except:
	return ''

# find the styles part of the rendered openoffice document
_stylere = re.compile(r'<style.*?>', re.DOTALL | re.I )
_endstylere = re.compile(r'</style', re.DOTALL | re.I )

def stylefinder( text ):

    style = _stylere.search( text )
    if not style:
	return style

    end = _endstylere.search( text )
    if not end:
	return text
    else:
	return text[style.end():end.start()]


class DynZipWrapper(Base):

    def __of__(self, parent):
        return parent.document()


class DynZip(Acquisition.Implicit, Traversable):
	""" temporary object to access the zipped file contents
            like images on demand
	"""
	def __init__(self, current=''):
	    self.current=current

        def getzip(self):
            return self.document(getzip=1)
		
        def read(self, name):
	    return self.getzip().read(name)

	def close(self):
	    return self.getzip().close()

	def __getitem__(self, name):
	    if self.current: spacer='/'
	    else: spacer=''
	    name='%s%s%s' % (self.current, spacer, name)
	    return DynZip(current=name).__of__(aq_parent(self))

	def __call__(self, *args, **kw):
	    """
	    """
	    REQUEST=self.REQUEST

            if not self.workingState() and self.hasImage(self.current):
                img = self.getImage(self.current)
                if img.getContentType().startswith('image'):
		    return img.__of__(aq_parent(self)).index_html(REQUEST, REQUEST.RESPONSE)

	    body=self.read(self.current)
	    fn=self.current.split('/')[-1]
	    content_type, enc=guess_content_type(fn, body)

	    REQUEST.RESPONSE.setHeader('Content-Type', content_type)
	    REQUEST.RESPONSE.setHeader('Content-Disposition', 'filename=%s' % (fn))
  	    return self.read(self.current)

	index_html=__call__

        def resize(self, quality=100):

	    path = self.current
	    link = "doc/%s" % (path)
            size = self.images_sizes.get(link, None)
	    if not size: return self.read(path)
	    width, height = size

            image=StringIO()

	    if isPilAvailable:
                img = PIL.Image.open(StringIO(str(self.read(path))))
                fmt = img.format
                # Resize image 
                img = img.resize((width, height), PIL.Image.BICUBIC)
                # Store copy in image buffer
                img.save(image, fmt, quality=quality)

	    else:
		return self.read(path)

	    image.seek(0)
	    return image.read()
 
factory_type_information = ( { 'id'             : 'OODocument'
                             , 'meta_type'      : 'OODocument'
                             , 'description'    : """\
OODocuments contain Documents written with OpenOffice Writer. The content is rendered as html."""
                             , 'icon'           : 'document_icon.gif'
                             , 'product'        : 'CMFOODocument'
                             , 'factory'        : 'addOODocument'
                             , 'immediate_view' : 'metadata_edit_form'
                             , 'actions'        :
                                ( { 'id'            : 'view' 
                                  , 'name'          : 'View'
                                  , 'action'        : 'document_view'
                                  , 'permissions'   : (View,)
                                  }
 				, { 'id'	    : 'download'
	                          , 'name'	    : 'Download'
				  , 'action'	    : 'file_view'
				  , 'permissions'   : (View,)
				  , 'visible'	    : 0
				  }
                                , { 'id'            : 'edit'
                                  , 'name'          : 'Edit'
                                  , 'action'        : 'file_edit_form'
                                  , 'permissions'   : (ModifyPortalContent,)
                                  }
                                , { 'id'            : 'metadata'
                                  , 'name'          : 'Metadata'
                                  , 'action'        : 'metadata_edit_form'
                                  , 'permissions'   : (ModifyPortalContent,)
                                  }
                                , { 'id'            : 'external_edit'
                                  , 'name'          : 'External Edit'
                                  , 'action'        : 'external_edit'
                                  , 'permissions'   : (ModifyPortalContent,)
				  , 'visible'	    : 0
                                  }
                                )
                             }
                           ,
                           )

def addOODocument( self
                 , id
		 , title=''
		 , file=''
		 , content_type=''
		 , precondition=''
		 , subject=() 
		 , description=''
		 , contributors=()
		 , effective_date=None
		 , expiration_date=None
		 , format=text_format
		 , language=''
		 , rights=''
                 ):
    """ Add a OODocument """

    # cookId sets the id and tile if they are not set
    id, title = OFS.Image.cookId(id, title, file)

    self=self.this()

    o = OODocument( id, title, '', content_type, precondition, subject
                   , description, contributors, effective_date, expiration_date
                   , format, language, rights
                   )
   
    # Add the OODocument instance to self
    self._setObject(id, o)

    # 'Upload' the file. This is done now rather than in the
    # contructor because the object is now in the ZODB and
    # can span ZODB objecs
    if not file: file=loadEmptyOODocument()
    self._getOb(id).edit(file=file)


class OODocument(File):
    """ A Document which holds and renders OpenOffice files to html """

    meta_type = 'OODocument'
    effective_date = expiration_date = None
    _isDiscussable = 1
    text='true' # this is set for compatibility with document_view
    text_format='html'
    css_styles=''
    images_sizes={}
    _images=None

    security = ClassSecurityInfo()

    security.declareProtected(View, 'index_html')
    def index_html(self, REQUEST=None, RESPONSE={}):
	"""
	Invokes the default view.
	"""
	view = _getViewFor(self, 'view', 'document_view')
	if getattr(aq_base(view), 'isDocTemp', 0):
	    return apply(view, (self, self.REQUEST))
	else:
	    return view()

    __call__=view=index_html

    security.declareProtected(View, 'CookedStyles')
    def CookedStyles(self, REQUEST=None, RESPONSE=None):
	"""
	returns the stylesheet
	"""
	if RESPONSE: RESPONSE.setHeader('Content-Type', 'text/css')
	return '\n'.join(('<style>'
                         ,getattr(self, 'css_styles', '')
                         ,'</style>'))

    security.declareProtected(View, 'CookedBody')
    def CookedBody(self, stx_level=None, setlevel=0):
	"""
	return openoffice file as html
	"""
	if getattr(self, 'keepOOStyles', 1):
	    styles="""
            <!-- openoffice styles for CSS2 capable browsers -->
            <style type="text/css"> @import url(%s/CookedStyles);</style>
	    """ % (self.absolute_url())
        else: styles=""

	return '\n'.join((styles
                         ,getattr(self, 'cooked_text', 'This OpenOffice Document has yet to be uploaded.')))

    security.declareProtected(View, 'SearchableBody')
    def SearchableBody(self):
	"""
	return html
	"""
	return getattr(self, 'cooked_text', '')

    security.declareProtected(View, 'Description')
    def Description(self):
	"""
	Dublin core description
	"""
	return self.description

    security.declareProtected(View, 'SearchableText')
    def SearchableText(self):
        """ Used by the catalog for basic full text indexing """
        return "%s %s %s" % ( self.Title()
                            , self.Description()
                            , self.SearchableBody()
                            )

    security.declareProtected(ModifyPortalContent, 'edit')
    def edit(self, precondition='', file=''):
	""" Update and reindex. """
	self._edit( precondition, file)
	self.reindexObject()

    security.declarePrivate('_edit')
    def _edit(self, precondition='', file=''):
        """ Edit the OODocument - Parses headers and cooks the body """

        if precondition: self.precondition = precondition
        elif self.precondition: del self.precondition

        if self._isNotEmpty(file):
            self.manage_upload(file)

        self.setFormat(text_format)

	# cook it
	self._cook()

    security.declarePrivate('_cook')
    def _cook(self):
	""" render the document by xslt to html """

	params={}
	# default params
	baseURL='doc'
	params.update({  'absoluteSourceDirRef': "'%s'" % (baseURL)
		       , 'disableJava': 'true()'
		       , 'disableLinkedTableOfContent': 'false()'
                       , 'dpi': "'86'"
		     })

	if _DEBUG_XSLT:
	    params.update({'DEBUG': 'true()'})

	# get the zipped document
	document=self.document()

	# create temp files
	tmpfiles=[]
	for f in (('stylesFileURL','styles.xml'), ('metaFileURL','meta.xml')):
	    # get unique filename and write contents to this file
            t=tempfile.mktemp()
            tmpfiles.append(t)
            file=open(t, 'wb', -1)
            file.write(document.read(f[1]))
            file.close()
	    # update the parameters to reflect the temporary files
            params.update({f[0]: "'%s'" % (t)})

	# read the content xml
	thisdoc=document.read('content.xml')
	# parse the content
	doc=libxml2.parseDoc(thisdoc)
	# apply the stylesheet (xslt transformation)
	result=_style.applyStylesheet(doc, params)
	# write html to string
	html=result.serialize(format=1)
	# free memory
	result.freeDoc()
	doc.freeDoc()
	# write rendered document to zodb
	self._storeCooked(body=html)
	# clean up created temp files
	for t in tmpfiles:
	    try: os.unlink(t)
	    except: pass


    security.declarePrivate('_parseImages')
    def _parseImages(self, html):
	""" parses all images out of the html code """
        regex = re.compile("(<img width=\"(\d+)\"[\s]+height=\"(\d+)\"[\s]+src=\"([\S]+)\")", re.M)

        matches={}
        def search(r, s):
	    return r.search(s)
        m = search(regex, html)
	if m: 
	    matches[m.group(4)]=(int(m.group(2)), int(m.group(3)))
  	    c=m.end()
	while(m):
	    html=html[c:]
	    m = search(regex, html)
	    if m: 
                c=m.end()
		matches[m.group(4)]=(int(m.group(2)), int(m.group(3)))

        return matches

    security.declareProtected(View, 'getImage')
    def getImage(self, path, default=None):
        '''returns the image of the specified zippath.'''
	if not self.hasImage(path): return default
        else: return self._images[path]

    security.declareProtected(View, 'hasImage')
    def hasImage(self, path):
        '''returns true if already there.'''
        if self._images is None: return 0
	try: return self._images.has_key(path)
	except Exception, e:
	    print "err", e
            return 0

    security.declarePrivate('setImage')
    def setImage(self, path, content):
	"""store image if not there."""
        if self._images is None: self._images=OOBTree()
        id=path.split('/')[-1]
	if not self.hasImage(path):
            self._images[path] = OFS.Image.Image(id, id, content)

    security.declarePrivate('_storeCooked')
    def _storeCooked(self, body=None):
	""" parse the rendered document and store metadata and body """
	if body == None: body=self.CookedBody()
	headers, body, styles = self.handleText(text=body)
        self.workingState(1)
        try:
            self.images_sizes=self._parseImages(body)
	    for path in self.images_sizes.keys():
                p=path[4:]
	        self.setImage(p, DynZip(p).__of__(self).resize())
        finally: self.workingState(0)
	self.cooked_text=body
        self.css_styles=styles
	self.setMetadata(headers)

    security.declarePrivate('workingState')
    def workingState(self, state=None):
        # set temp flag to indicate longer document processing
        # to avoid conflicts while image scaling
	if state is None: return getattr(self, '_v_workingState', 0)
        else: 
	    self._v_workingState=state
            return state


    security.declarePrivate('handleText')
    def handleText(self, text):
	""" Handles the raw html content returning headers and body """
	headers = {}
        styles = ''
	# create html parser
	parser = SimpleHTMLParser()
	# parse for headers and body
	parser.feed(text)
	# update headers dictionary
	headers.update(parser.metatags)
	if parser.title:
	    headers['Title'] = parser.title
	# parse the styles 
        stylesfound = stylefinder(text)
	if stylesfound: styles=stylesfound
	# strip body
        bodyfound = bodyfinder(text)
	if bodyfound: text=bodyfound
	# return data
	return headers, text, styles

    security.declarePrivate('update_data')
    def update_data(self, data, content_type=None, size=None):
    	#print "content_type", content_type
        if content_type is not None: self.content_type=text_format
        if size is None: size=len(data)
	self._images = OOBTree()
        self.size=size
        self.data=data
        self.ZCacheable_invalidate()
        self.http__refreshEtag()

    security.declareProtected(View, 'document')
    def document(self, getzip=0):
	"""
	returns a dynzip instance
	"""
        if not getzip:
	    return DynZip().__of__(self)

	# XXX: how is that zipfile and stringio instance closed?
	if type(self.data) != type('') and type(self.data) != type(u'') and isinstance(self.data, OFS.Image.Pdata):
	    # if things get big we have Pdata to deal with
	    c=self.data
	    data=StringIO()
	    next=c.next
	    data.write(c.data)
	    while next is not None:
		c=next
		data.write(c.data)
		next=c.next
	else:
	    # else data has to be something which can be transformed to StringIO
	    data=StringIO(self.data)		

	# create zip extraction context
	return zipfile.ZipFile(data, 'rb')

    # make zip contents accessable via web
    doc=DynZipWrapper()

    security.declareProtected(ModifyPortalContent, 'setMetadata')
    setMetadata = Document.setMetadata

    security.declareProtected(ModifyPortalContent, 'PUT')
    def PUT(self, REQUEST, RESPONSE):
        """ Handle HTTP PUT requests """
        self.dav__init(REQUEST, RESPONSE)
        self.dav__simpleifhandler(REQUEST, RESPONSE, refresh=1)
        type=REQUEST.get_header('content-type', None)

        file=REQUEST['BODYFILE']

        data, size = self._read_data(file)
        content_type=self._get_content_type(file, data, self.__name__,
                                            type or self.content_type)
        self.update_data(data, content_type, size)
	self._cook()

        RESPONSE.setStatus(204)
        return RESPONSE



def _getViewFor(obj, view='view', default=None):
    ti = obj.getTypeInfo()
    if ti is not None:
        actions = ti.getActions()
        for action in actions:
            if action.get('id', None) == default:
                default=action
            if action.get('id', None) == view:
                if _verifyActionPermissions(obj, action) and action['action']!='':
                    return obj.restrictedTraverse(action['action'])

        if default is not None:
            if _verifyActionPermissions(obj, default):
                return obj.restrictedTraverse(default['action'])

        raise 'Unauthorized', ('No accessible views available for %s' %
                               '/'.join(obj.getPhysicalPath()))
    else:
        raise 'Not Found', ('Cannot find default view for "%s"' %
                            '/'.join(obj.getPhysicalPath()))

InitializeClass(OODocument)
